"""
This script is based on the solution of exercise 03.2.
"""

from scipy.integrate import solve_ivp
from scipy.optimize import minimize
import matplotlib.pyplot as plt
import numpy as np

# import of the functions for the right hand side of the model
from equations_of_motion import xdd_fnc, phidd_fnc


# Reference dataset to which the model is to be optimized.
# Load data in binary format
measurment_data = np.load("measurement-data.npy")

# array with evaluation times (should be consistent with the measured data)
tt = np.linspace(0, 10, 1001)

zz0 = measurment_data[:, 0] # first column of measurement data -> initial value of the simulation

# Create empty list for collecting intermediate optimization results:
resList = []


# this is copied from exercise 03.2 (+ one line added)
def min_target(p):
    """
    This function is called repeatedly by the optimization algorithm.
    On every call it performs an simlation and calculates the difference
    to the measurement data.

    :param  p:  sequence of two parameter values (m2, l)

    :returns:   err – non-negative real valued error measure
                (how "wrong the simulation result is")
    """
    m2, l = p # unpacking the parameter vector

    # Define a function INSIDE another function.
    # works and is often useful!
    # The inner function has reading access on the namespace
    # of the outer function (as with global variables).

    # The outer function can not access the namespace of the inner function.

    def rhs(t, z):
        """
        Righthand side of the equations of motion
        (Note: this depends on m2 and l from the surrounding namespace).
        """
        x, phi, xd, phid = z # unpacking
        F = 0

        # m2 and l come from the namespace one level higher
        # (you might want to look again into `eqations_of_motion.py`
        # to check the signature of these functions:)
        xdd = xdd_fnc(x, phi, xd, phid, F, m2, l)
        phidd = phidd_fnc(x, phi, xd, phid, F, m2, l)

        # return derivative of the state vector
        return np.array([xd, phid, xdd, phidd])

    # end of the inner function definition of rhs

    # do the simulation (get result container)
    sim_res = solve_ivp(rhs, (tt[0], tt[-1]), zz0, t_eval=tt, rtol=1e-5)

    # select the state vector (which we call "z" but scipy calls "y")
    zz_res = sim_res.y




    # Exercise 03.1:
    # That single line is new:
    # save the current simulation result (state array) in the prepared list
    resList.append(zz_res)




    # Calculate the difference of the x-positions (first line in each case)
    # then square (...**2),
    # then add up (applying np.sum)
    err = np.sum( (zz_res[0, :] - measurment_data[0, :])**2 )

    # Status message and output (to assess progress of optimization)
    print("simulation ready. p =", p, " equation error:", err)

    return err # return scalar (real valued) error measure


## To save time in case of repeated execution (e.g. to adjust the plots).
## you should save the result data of the simulation.
## Here the serialization module pickle is used.
## With this module (almost) arbitrary Python objects can be stored in a file.
## See https://docs.python.org/3/library/pickle.html

## Alternatively, for numeric data, of course, you can use the
## functions numpy.save and numpy.load.

## In general, a separation of (complex) data generation and
## visualization is often very useful.
## Here this separation is achieved by an if-block which can easily
# be switchend on and of. (if 0: ... -> if 1: ...).



import pickle # module for serialization
pfname = "res_list.pcl" # arbitrary file name


## After successful saving, insert a 0 here:
if 0:
    # Optimization - result are the two parameters m2 and l

    # intially guessed values for optimization
    p0 = [.5, .7]

    # Carry out the optimization
    p_res = minimize(min_target, p0, method="Nelder-Mead")

    # open file for writing in binary mode
    with open(pfname, "wb") as myfile:
        pickle.dump(resList, myfile) # resList is serialized and stored
    print("Simulation data successfully saved:", pfname)
    print("Program end")
    exit()
else:
    # open file for reading in binary mode
    with open(pfname, "rb") as myfile:
        # load calculation results
        resList = pickle.load(myfile)


#-------------------------------------------------------------------------------------
# Exercise 03.1: Plot measured values, optimization result and error (3x1 arrangement)
#-------------------------------------------------------------------------------------

mm = 1./25.4 # scaling factor for mm -> inch
fig = plt.figure(figsize=(300*mm, 170*mm))

# subplot for unoptimized curves
ax1 = fig.add_subplot(3, 1, 1)
ax1.plot(tt, measurment_data[0, :], label="measurement")
ax1.plot(tt, resList[0][0, :], label="model (initial guess)")
ax1.legend()
ax1.grid()
ax1.set_ylabel("distance $x ~[m]$")


# subplot for optimization result
ax2 = fig.add_subplot(3, 1, 2)
ax2.plot(tt, measurment_data[0, :], lw=3, label="measurement")
ax2.plot(tt, resList[-1][0, :], label="model (optimized parameters)")
ax2.legend()
ax2.grid()
ax2.set_ylabel("distance $x ~[m]$")


# subplot for remaining error of x position
ax3 = fig.add_subplot(3, 1, 3)
ax3.plot(tt, measurment_data[0, :]-resList[-1][0, :])
ax3.grid()
ax3.set_xlabel("time $t ~[s]$")
ax3.set_ylabel("error $\epsilon$ ~[m]")

plt.savefig("fig1.pdf", bbox_inches="tight")
# plt.show()


#-------------------------------------------------------------------------------
# Exercise 03.2: Visualize optimization progress (2d)
#-------------------------------------------------------------------------------

fig = plt.figure(figsize=(300*mm, 170*mm))

# Plot all intermediate results in different shades of gray (0.1 < col < 0.8)
for i in range( len(resList) - 1 ):
    colStep = 0.9 / len(resList)
    gray_value = str(0.01 + (0.9 - i*colStep))

    plt.plot(tt, resList[i][0, :], color=gray_value )

# Plotting the measurement and the simulation
plt.plot(tt, measurment_data[0, :], color="#3366FF", lw=4, label="measurement")
plt.plot(tt, resList[-1][0, :], color="#FF9900", ls="--", lw=2, label="model (optimized parameters)")
plt.grid()
plt.legend()
plt.xlabel("time $t ~[s]$")
plt.ylabel("distance $x ~[m]$")

# here we use png because due to the many individual lines the pdf would be quite big and slow
plt.savefig("fig2.png", bbox_inches="tight")
# plt.show()

#-------------------------------------------------------------------------------
# Exercise 03.2: Visualize optimization progress (3d)
#-------------------------------------------------------------------------------

from mpl_toolkits.mplot3d import axes3d
from matplotlib import cm
from pylab import meshgrid

fig = plt.figure()
ax = fig.add_subplot(1, 1, 1, projection="3d")

# This is a "list comprehension" (will be explained in ≈ course06)
x_res = [i[0, :] for i in resList]

Z = np.vstack(x_res).T
X, Y = meshgrid(range(len(resList)), tt)

# overview on color maps:
# https://matplotlib.org/examples/color/colormaps_reference.html
ax.plot_surface(X, Y, Z, rstride=1, cstride=1, cmap=cm.viridis)
ax.set_xlabel("x")
ax.set_ylabel("y")
ax.set_zlabel("z")

# here we use png because due to the many individual lines the pdf would be quite big and slow
plt.savefig("fig3.png", bbox_inches="tight")
plt.show()
