import numpy as np
from numpy import sin, cos
import vtk


def calcPositions(y):
    """
    Calculate the position and orientation of bodies in space (in a global coordinate system).
    Hereby y is the current state vector from numerical integration ([x, phi, xd, phid]).
    (the argument `y` is not to be confused with the direction "y" in 3D coordinates (x, y, z) )

    Also, we have variables with the following meanings:
    r - postion (array with shape (3,) i.e. 3D vector)
    R - orientation (array with shape (3, 3), i.e. a 3x3 matrix)

    This functions returns a tuple with r- and R-values for trolley, load (and rope)
    """

    # length of the pendulum
    l = 0.5

    # extract trolley position x and pendulum angle phi from the state vector
    # (the velocity components are not needed)
    x = y[0]
    phi = y[1]

    # calculate position and orientation of the trolley
    r_trolley = np.array([x, 0, 0])
    R_trolley = np.eye(3)  # no orientation change needed (use unit matrix)

    # calculate position and orientation of the load
    r_load = np.array([x+l*sin(phi), 0, -l*cos(phi)])

    # default rotation matrix about y-axis
    R_load = np.array([[cos(phi), 0, -sin(phi)],
                       [          0, 1, 0           ],
                       [sin(phi), 0,  cos(phi)]])


    # task 9:

    # calculate position and orientation of the rope
    # Zylinder muss zusätzlich um 90 Grad um x-Achse in die x-z-Ebene gedreht werden
    r_rope = np.array([x+l/2*sin(phi),0,-l/2*cos(phi)])

    # the cylinder that represents the rope has to be rotated also about the x-axis
    # this can be achieven by usinge the according default matrix:
    Rx = np.array([[1,            0,           0],
                   [0,  cos(np.pi/2), sin(np.pi/2)],
                   [0, -sin(np.pi/2), cos(np.pi/2)]])

    # the orientation matrix of the rope is now the orientation matrix of the
    # load (depending on phi) multiplied with that static rotation (Rx)
    R_rope = R_load@Rx

    # return a tuple with 6 elements (brackets are optional)
    return (r_trolley, R_trolley, r_load, R_load, r_rope, R_rope)


def setPokeMatrix(actor, r, R):
    """
    The position of 3D objects in vtk is defined with a so-called poke matrix.
    It is a 4x4 matrix with the following shape:

    [          ]
    [   R     r]
    [          ]
    [0  0  0  1]

    If you want to update the position and orientation of a body, the values of the
    poke matrix must be changed. This is done with the help of this function.

    :param actor:   vtk Actor instance
    :param r:       array with shape (3,) – displacement vector
    :param R:       array with shape (3, 3) – orientation matrix
    """

    # create empty vtk matrix (actually is a 4x4 unit matrix)
    # its elements will be overwritten soon
    poke = vtk.vtkMatrix4x4()

    # stack together the received arguments of the functions
    P = np.column_stack((R, r))
    P = np.row_stack( (P, np.array([0, 0, 0, 1])) )

    # P is now a numpy array with shape (4, 4); its elements are copied to the above vtk matrix
    vtk.vtkMatrix4x4.DeepCopy(poke, P.flatten())

    # pass the poke matrix to the actor
    actor.PokeMatrix(poke)
